class PathNotExistsError(Exception):
    def __init__(self, path: str) -> None:
        super().__init__()
        self.args = (f"The element along the path {repr(path)} does not exist.",)

class NotBooleanError(Exception):
    def __init__(self, data: str) -> None:
        super().__init__()
        self.args = (f"The data {repr(data)} cannot be represented as a bool.",)


import time
from textual.widgets import Label
# > Image Works
from PIL import Image
from PIL.Image import Resampling
from ripix import AsyncPixels, Pixels
# > Typing
from typing import Optional, Union, Tuple


class StandartImageLabel(Label):
    def __init__(
        self,
        default_image: Image.Image,
        image: Optional[Image.Image]=None,
        *,
        resample: Resampling=Resampling.NEAREST
    ):
        super().__init__("<image not found>", classes="image-label")
        self.image_resample = resample
        self.default_image: Image.Image = default_image
        self.image: Optional[Image.Image] = image
        self.image_text: Union[str, Pixels] = "<image not found>"
        self.last_image_size: Optional[Tuple[int, int]] = None
    
    async def on_resize(self) -> None:
        image, resample = (self.default_image, Resampling.NEAREST) if (self.image is None) else (self.image, self.image_resample)
        new_size = (self.size[0], self.size[1])
        if self.last_image_size != new_size:
            self.image_text = Pixels.from_image(image, new_size, resample)
            self.last_image_size = new_size
        self.update(self.image_text)
    
    async def update_image(self, image: Optional[Image.Image]=None) -> None:
        self.image = image
        
        image, resample = (self.default_image, Resampling.NEAREST) if (self.image is None) else (self.image, self.image_resample)
        self.image_text = Pixels.from_image(image, (self.size[0], self.size[1]), resample)
        self.update(self.image_text)

class AsyncImageLabel(Label):
    def __init__(
        self,
        default_image: Image.Image,
        image: Optional[Image.Image]=None,
        *,
        resample: Resampling=Resampling.NEAREST
    ):
        self.image_resample = resample
        self.default_image: Image.Image = default_image
        self.image: Optional[Image.Image] = image
        self.image_text: Union[str, AsyncPixels] = "<image not found>"
        self.last_image_size: Optional[Tuple[int, int]] = None
        super().__init__("<image not found>", classes="image-label")
    
    async def on_resize(self) -> None:
        image, resample = (self.default_image, Resampling.NEAREST) if (self.image is None) else (self.image, self.image_resample)
        new_size = (self.size[0], self.size[1])
        if self.last_image_size != new_size:
            self.image_text = await AsyncPixels.from_image(image, new_size, resample)
            self.last_image_size = new_size
        self.update(self.image_text)
    
    async def update_image(self, image: Optional[Image.Image]=None) -> None:
        self.image = image
        
        image, resample = (self.default_image, Resampling.NEAREST) if (self.image is None) else (self.image, self.image_resample)
        self.image_text = await AsyncPixels.from_image(image, (self.size[0], self.size[1]), resample)
        self.update(self.image_text)


from pydantic import BaseModel
# > Typing
from typing import Optional
# > Local Import's
from ..seaplayer import SeaPlayer
from .pluginloader import PluginLoader

# ! Plugin Info Class
class PluginInfo(BaseModel):
    name: str
    name_id: str
    version: str
    author: str
    description: Optional[str]=None
    url: Optional[str]=None

# ! Plugin Base Class
class PluginBase:
    app: SeaPlayer
    pl: PluginLoader
    info: PluginInfo
    
    def __init__(
        self,
        app: SeaPlayer,
        pl: PluginLoader,
        info: PluginInfo
    ) -> None: ...

    def on_init(self) -> None: ...
    def on_run(self) -> None: ...
    async def on_compose(self) -> None: ...
    async def on_quit(self) -> None: ...



from textual.widgets import Input
# > Typing
from typing import Optional, Tuple, Any

# ! InputFiles functions
async def _conv(value: str) -> Tuple[bool, Optional[Any]]: return True, value
async def _submit(input: Input, value: Any) -> None: ...
def _update_placeholder() -> Optional[str]: ...

# ! InputField class
class InputField(Input):
    def __init__(
        self,
        conv=_conv,
        submit=_submit,
        update_placeholder=_update_placeholder,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self._conv = conv
        self._submit = submit
        self._update_placeholder = update_placeholder
        if (placeholder:=self._update_placeholder()) is not None:
            self.placeholder = placeholder
    
    async def action_submit(self):
        value = self.value
        self.value = ""
        if value.replace(" ", "") != "":
            ok, c_value = await self._conv(value)
            if ok: await self._submit(self, c_value)
        if (placeholder:=self._update_placeholder()) is not None:
            self.placeholder = placeholder


from rich.progress import Progress, BarColumn, TextColumn
from textual.widgets import Static
# > Typing
from typing import Optional
# > Local Import's
from ..functions import get_bar_status


class IndeterminateProgress(Static):
    def __init__(self, getfunc=get_bar_status, fps: int=15):
        super().__init__("", classes="indeterminate-progress-bar")
        self._bar = Progress(BarColumn(), TextColumn("{task.description}"))
        self._task_id = self._bar.add_task("", total=None)
        self._fps = fps
        self._getfunc = getfunc
    
    def on_mount(self) -> None:
        self.update_render = self.set_interval(1/self._fps, self.update_progress_bar)
    
    async def upgrade_task(self, description: str="", completed: Optional[float]=None, total: Optional[float]=None) -> None:
        self._bar.update(self._task_id, total=total, completed=completed, description=description)
    
    async def update_progress_bar(self) -> None:
        d, c, t = await self._getfunc()
        if self._bar.columns[0].bar_width != (self.size[0]-len(d)-1):
            self._bar.columns[0].bar_width = self.size[0]-len(d)-1
        
        await self.upgrade_task(completed=c, total=t, description=d)
        self.update(self._bar)



from textual.widgets import Label, ListItem, ListView
# > Typing
from typing import Optional
# > Local Import's
from ..types import MusicList
from ..codeсbase import CodecBase
from ..functions import get_sound_basename, aiter


class MusicListViewItem(ListItem):
    def __init__(
        self,
        title: str="",
        first_subtitle: str="",
        second_subtitle: str="",
        sound_uuid: Optional[str]=None
    ) -> None:
        super().__init__(classes="music-list-view-item")
        self.title_label = Label(title, classes="music-list-view-item-title-label")
        self.first_subtitle_label = Label(f" {first_subtitle}", classes="music-list-view-item-subtitle-label")
        self.second_subtitle_label = Label(f" {second_subtitle}", classes="music-list-view-item-subtitle-label")
        self.sound_uuid = sound_uuid
        
        self.compose_add_child(self.title_label)
        self.compose_add_child(self.first_subtitle_label)
        self.compose_add_child(self.second_subtitle_label)
    
    async def update_labels(
        self,
        title: Optional[str]=None,
        first_subtitle: Optional[str]=None,
        second_subtitle: Optional[str]=None,
    ) -> None:
        if title is not None: self.title_label.update(title)
        if first_subtitle is not None: self.first_subtitle_label.update(title)
        if second_subtitle is not None: self.second_subtitle_label.update(title)

class MusicListView(ListView):
    def __init__(self, **kwargs) -> None:
        kwargs["classes"] = "music-list-view"
        super().__init__(**kwargs)
        self.music_list: MusicList = MusicList()
    
    def add_sound(self, sound: CodecBase) -> str:
        sound_uuid = self.music_list.add(sound)
        self.append(
            MusicListViewItem(
                f"{get_sound_basename(sound)}",
                "{duration} sec, {channel_mode}, {samplerate} Hz, {bitrate} kbps".format(
                    duration=round(sound.duration),
                    channel_mode="Mono" if sound.channels <= 1 else "Stereo",
                    samplerate=round(sound.samplerate),
                    bitrate=round(sound.bitrate / 1000)
                ),
                sound.name,
                sound_uuid
            )
        )
        return sound_uuid
    
    def get_sound(self, sound_uuid: str) -> Optional[CodecBase]: return self.music_list.get(sound_uuid)
    
    async def aio_add_sound(self, sound: CodecBase):
        sound_uuid = await self.music_list.aio_add(sound)
        await self.append(
            MusicListViewItem(
                f"{get_sound_basename(sound)}",
                "{duration} sec, {channel_mode}, {samplerate} Hz, {bitrate} kbps, {codec_name}".format(
                    duration=round(sound.duration),
                    channel_mode="Mono" if sound.channels <= 1 else "Stereo",
                    samplerate=round(sound.samplerate),
                    bitrate=round(sound.bitrate / 1000),
                    codec_name=sound.codec_name
                ),
                sound.name,
                sound_uuid
            )
        )
        return sound_uuid
    
    def get_items_count(self) -> int: return len(self.children)
    def exists_item_index(self, index: int) -> bool: return 0 >= index < self.get_items_count()
    def get_item_index_from_sound_uuid(self, sound_uuid: str) -> Optional[int]:
        item: MusicListViewItem
        for idx, item in enumerate(self.children):
            if item.sound_uuid == sound_uuid:
                return idx
    
    def get_item_from_index(self, index: int) -> Optional[MusicListViewItem]:
        try: return self.children[index]
        except:
            try: return self.children[0]
            except: pass
    
    def get_next_sound_uuid(self, sound_uuid: str) -> Optional[str]:
        if (index:=self.get_item_index_from_sound_uuid(sound_uuid)) is not None:
            if (mli:=self.get_item_from_index(index+1)) is not None:
                return mli.sound_uuid
    
    def select_list_item_from_sound_uuid(self, sound_uuid: str) -> None:
        try:
            super()._on_list_item__child_clicked(
                ListItem._ChildClicked(
                    self.children[self.get_item_index_from_sound_uuid(sound_uuid)]
                )
            )
        except: pass
    
    async def aio_get_item_from_index(self, index: int) -> Optional[MusicListViewItem]:
        try: return self.children[index]
        except:
            try: return self.children[0]
            except: pass
    
    async def aio_get_item_index_from_sound_uuid(self, sound_uuid: str) -> Optional[int]:
        item: MusicListViewItem
        async for idx, item in aiter(enumerate(self.children)):
            if item.sound_uuid == sound_uuid:
                return idx
    
    async def aio_get_next_sound_uuid(self, sound_uuid: str) -> Optional[str]:
        if (index:=await self.aio_get_item_index_from_sound_uuid(sound_uuid)) is not None:
            if (mli:=await self.aio_get_item_from_index(index+1)) is not None:
                return mli.sound_uuid
    
    async def aio_select_list_item_from_sound_uuid(self, sound_uuid: str) -> None:
        try:
            super()._on_list_item__child_clicked(
                ListItem._ChildClicked(
                    self.children[await self.aio_get_item_index_from_sound_uuid(sound_uuid)]
                )
            )
        except: pass


from typing import Optional

# ! Functions
def formater(**kwargs) -> str:
    return ", ".join([f"{key}={repr(value)}" for key, value in kwargs.items()])

# ! Codecs Base
class CodecBase:
    # * Codec Info
    codec_name: str = "None"
    
    # * Info
    name: str
    duration: float
    channels: int
    samplerate: int
    bitrate: int
    
    # * Sound Info
    title: Optional[str]
    artist: Optional[str]
    album: Optional[str]
    icon_data: Optional[bytes]
    
    # * Playback Info
    playing: bool
    paused: bool
    
    # ! Initializing Functions
    def __init__(self, path: str, **kwargs) -> None: ...
    #async def __aio_init__(self, path: str, **kwargs) -> None: ...
    def __repr__(self):
        return \
            "{0}({1})".format(
                self.__class__.__name__,
                formater(
                    name=self.name,
                    duration=self.duration,
                    channels=self.channels,
                    samplerate=self.samplerate,
                    bitrate=self.bitrate,
                    title=self.title,
                    artist=self.artist,
                    album=self.album
                )
            )
    
    def __sha1__(self, buffer_size: int) -> str: ...
    async def __aio_sha1__(self, buffer_size: int) -> str: ...
    
    # ! Testing Functions
    @staticmethod
    def is_this_codec(path: str) -> bool: return False
    @staticmethod
    async def aio_is_this_codec(path: str) -> bool: return False
    
    # ! Playback Functions
    def play(self) -> None: ...
    def stop(self) -> None: ...
    def pause(self) -> None: ...
    def unpause(self) -> None: ...
    def get_volume(self) -> float: return 1.0
    def set_volume(self, value: float) -> None: ...
    def get_pos(self) -> float: return 0.0
    def set_pos(self, value: float) -> None: ...



from pathlib import Path
from pydantic import BaseModel
# > Typing
from types import ModuleType
from typing import (
    Any,
    Dict,
    List,
    Type,
    Tuple,
    Union,
    Optional,
    Generator, 
    AsyncGenerator
)
# > Local Import's
from .pluginbase import PluginBase, PluginInfo
from ..seaplayer import SeaPlayer

# ! Types
class PluginModuleType(ModuleType):
    plugin_main: Type[PluginBase]

# ! Functions
def get_module_info(path: str) -> Tuple[str, str]: ...
def load_module(path: str) -> PluginModuleType: ...
def plugin_from_module(app: SeaPlayer, pl: PluginLoader, info: PluginInfo, module: PluginModuleType) -> PluginBase: ...

# ! Plugin Loader Config
class PluginLoaderConfigModel(BaseModel):
    plugins_enable: Dict[str, bool] = {}

class PluginLoaderConfigManager:
    filepath: Path
    config: PluginLoaderConfigModel
    
    @staticmethod
    def dump(path: str, data: PluginLoaderConfigModel) -> None: ...
    @staticmethod
    def load(path: str, default_data: Dict[str, Any]) -> PluginLoaderConfigModel: ...
    def refresh(self) -> None: ...

    def __init__(self, path: str) -> None: ...

    def exists_plugin(self, info: PluginInfo) -> bool: ...
    def add_plugin(self, info: PluginInfo) -> None: ...
    def is_enable_plugin(self, info: PluginInfo) -> bool: ...

# ! Plugin Loader Class
class PluginLoader:
    app: SeaPlayer
    plugins_dirpath: Path
    plugins_config_path: Path
    on_plugins: List[PluginBase]
    off_plugins: List[PluginInfo]
    error_plugins: List[Tuple[str, str]]

    def __init__(
        self,
        app: SeaPlayer,
        plugins_dirpath: Optional[Union[str, Path]]=None,
        plugins_config_path: Optional[Union[str, Path]]=None,
        *args,
        **kwargs
    ) -> None:
        ...
    
    async def aio_search_plugins_paths(self) -> AsyncGenerator[Tuple[str, str]]: ...
    def search_plugins_paths(self) -> Generator[Tuple[str, str], Any, None]: ...

    def on_init(self) -> None: ...
    def on_run(self) -> None: ...
    async def on_compose(self) -> None: ...
    async def on_quit(self) -> None: ...
    


from textual.containers import ScrollableContainer, Container
# > Typing
from typing import Optional


class ConfigurateListItem(Container):
    def __init__(
        self,
        *children,
        title: str="",
        desc: str="",
        **kwargs
    ):
        kwargs["classes"] = "configurate-list-view-item"
        super().__init__(*children, **kwargs)
        self.border_title = title
        self.border_subtitle = desc
    
    async def updating(self, title: Optional[str]="", desc: Optional[str]="") -> None:
        if title is not None: self.border_title = title
        if desc is not None: self.border_subtitle = desc

class ConfigurateList(ScrollableContainer):
    def __init__(self, *children, **kwargs):
        kwargs["classes"] = "configurate-list-view"
        super().__init__(*children, **kwargs)
        self.border_title = "Configurate"


import os
import asyncio
import aiofiles
import subprocess
from tempfile import mkstemp
# > Sound Works
from playsoundsimple import Sound
from playsoundsimple.units import SOUND_FONTS_PATH, FLUID_SYNTH_PATH
from playsoundsimple.exceptions import FileTypeError
from playsoundsimple.player import SoundFP, get_sound_filepath
# > Typing Import
from typing import Optional
# > Local Imports
from .Any import AnyCodec


# ! Codec Types
class MIDISound(Sound):
    async def aio_from_midi(
        fp: SoundFP,
        sound_fonts_path: Optional[str]=None,
        **kwargs
    ):
        path, is_temp = get_sound_filepath(fp, filetype=".midi")
        sound_fonts_path = sound_fonts_path or SOUND_FONTS_PATH
        if path is None: raise FileTypeError(fp)
        npath = mkstemp(suffix=".wav")[1]
        
        process = await asyncio.create_subprocess_exec(
            FLUID_SYNTH_PATH, "-ni", sound_fonts_path, path, "-F", npath, "-q",
            stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        await process.wait()
        
        if is_temp:
            try: os.remove(path)
            except: pass
        
        return Sound(npath, **{"is_temp": True, **kwargs})
    
    def from_midi(
        fp: SoundFP,
        sound_fonts_path: Optional[str]=None,
        **kwargs
    ):
        path, is_temp = get_sound_filepath(fp, filetype=".midi")
        sound_fonts_path = sound_fonts_path or SOUND_FONTS_PATH
        if path is None: raise FileTypeError(fp)
        npath = mkstemp(suffix=".wav")[1]
        
        subprocess.call(
            [FLUID_SYNTH_PATH, "-ni", sound_fonts_path, path, "-F", npath, "-q"],
            stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        
        if is_temp:
            try: os.remove(path)
            except: pass
        
        return Sound(npath, **{"is_temp": True, **kwargs})

# ! Codec
class MIDICodec(AnyCodec):
    codec_name: str = "MIDI"
    
    # ! Testing
    @staticmethod
    def is_this_codec(path: str) -> bool:
        with open(path, 'rb') as file:
            return file.read(4) == b"MThd"
    
    @staticmethod
    async def aio_is_this_codec(path: str) -> bool:
        async with aiofiles.open(path, 'rb') as file:
            return await file.read(4) == b"MThd"
    
    # ! Initialized
    def __init__(self, path: str, aio_init: bool=False, **kwargs) -> None:
        self.name = os.path.abspath(path)
        if not aio_init:
            self._sound = MIDISound.from_midi(self.name, **kwargs)
    
    @staticmethod
    async def __aio_init__(path: str, **kwargs):
        self = MIDICodec(path, aio_init=True)
        self._sound = await MIDISound.aio_from_midi(self.name, **kwargs)
        return self


from .seaplayer import SeaPlayer
from rich.console import Console

console = Console()

# ! Run Functions
def run() -> None:
    try:
        app = SeaPlayer()
        app.run()
        app.started = False
    except:
        console.print_exception(word_wrap=True, show_locals=True)

# ! Start
if __name__ == "__main__":
    run()


import aiofiles
# > Local Imports
from .Any import AnyCodec


class OGGCodec(AnyCodec):
    codec_name: str = "OGG"
    
    # ! Testing
    @staticmethod
    def is_this_codec(path: str) -> bool:
        with open(path, "rb") as file:
            return file.read(4) == b'OggS'
    
    @staticmethod
    async def aio_is_this_codec(path: str) -> bool:
        async with aiofiles.open(path, "rb") as file:
            return await file.read(4) == b'OggS'



from typing import Dict
from typing_inspect import (
    is_optional_type,
    is_union_type,
    is_literal_type,
    is_tuple_type,
    get_args,
    get_origin
)

# ! Vars
REPLACE_TYPES = {
    "str": "[green]str[/]",
    "bool": "[green]bool[/]",
    "int": "[green]int[/]",
    "float": "[green]float[/]",
    "bytes": "[green]bytes[/]",
    "bytearray": "[green]bytearray[/]",
    "complex": "[green]complex[/]",
    "list": "[green]list[/]",
    "tuple": "[green]tuple[/]",
    "dict": "[green]dict[/]",
    "None": "[cyan]None[/]",
    "Tuple": "[green]Tuple[/]",
    "Dict": "[green]Dict[/]",
    "List": "[green]List[/]",
    "Any": "[white]Any[/]",
    "Literal": "[green]Literal[/]",
    "['": "[[yellow]'",
    ", '": ", [yellow]'",
    "']": "'[/]]",
    "', ": "'[/], "
}

# ! Other Functions
def is_list_type(tp) -> bool: return get_origin(tp) == list
def is_dict_type(tp) -> bool: return get_origin(tp) == dict
def replaces(string: str, replacement: Dict[str, str]) -> str:
    for _old, _new in replacement.items(): string = string.replace(_old, _new)
    return string

# ! Functions
def pullyper(tp: type) -> str:
    if tp is None: return "None"
    elif is_optional_type(tp):
       return f"{pullyper(get_args(tp)[0])} | None"
    elif is_union_type(tp):
        return " | ".join(pullyper(arg) for arg in get_args(tp))
    elif is_literal_type(tp):
        return f"Literal[{', '.join(repr(arg) for arg in get_args(tp))}]"
    elif is_tuple_type(tp):
        if len(args:=get_args(tp)) > 0: return f"Tuple[{', '.join(pullyper(arg) for arg in args)}]"
        return "Tuple"
    elif is_list_type(tp):
        if len(args:=get_args(tp)) > 0: return f"List[{pullyper(args[0])}]"
        return "List"
    elif is_dict_type(tp):
        if len(args:=get_args(tp)) > 0: return f"Dict[{', '.join(pullyper(arg) for arg in args)}]"
        return "Dict"
    return tp.__name__

def richefication(tp: type) -> str: return replaces(pullyper(tp), REPLACE_TYPES)


import os
import hashlib
import aiofiles
# > Sound Works
from playsoundsimple import Sound
# > Typing
from typing import Optional
# > Local Imports
from ..codeсbase import CodecBase


class AnyCodec(CodecBase):
    codec_name: str = "Any"
    
    # ! Initialized
    def __init__(self, path: str, **kwargs) -> None:
        self.name = os.path.abspath(path)
        self._sound = Sound(self.name)
    
    def __sha1__(self, buffer_size: int) -> str:
        sha1 = hashlib.sha1()
        with open(self.name, "rb") as file:
            while True:
                data = file.read(buffer_size)
                if not data: break
                sha1.update(data)
        return sha1.hexdigest()
    
    async def __aio_sha1__(self, buffer_size: int) -> str:
        sha1 = hashlib.sha1()
        async with aiofiles.open(self.name, "rb") as file:
            while True:
                data = await file.read(buffer_size)
                if not data: break
                sha1.update(data)
        return sha1.hexdigest()
    
    # ! Info
    @property
    def duration(self) -> float: return self._sound.duration
    @property
    def channels(self) -> int: return self._sound.channels
    @property
    def samplerate(self) -> int: return self._sound.samplerate
    @property
    def bitrate(self) -> int: return self._sound.bitrate
    
    # ! Playback Info
    @property
    def playing(self) -> bool: return self._sound.playing
    @property
    def paused(self) -> bool: return self._sound.paused
    
    # ! Sound Info
    @property
    def title(self) -> Optional[str]: return self._sound.title
    @property
    def artist(self) -> Optional[str]: return self._sound.artist
    @property
    def album(self) -> Optional[str]: return self._sound.album
    @property
    def icon_data(self) -> Optional[bytes]: return self._sound.icon_data
    
    # ! Functions
    def play(self) -> None: self._sound.play()
    def stop(self) -> None: self._sound.stop()
    def pause(self) -> None: self._sound.pause()
    def unpause(self) -> None: self._sound.unpause()
    def get_volume(self) -> float: return self._sound.get_volume()
    def set_volume(self, value: float) -> None: self._sound.set_volume(value)
    def get_pos(self) -> float: return self._sound.get_pos()
    def set_pos(self, value: float) -> None: self._sound.set_pos(value)


import base64
import platform
# > Graphics
from textual.app import ComposeResult
from textual.screen import Screen
from textual.widgets import Static

# ! Constants
TEXT = b'CkFuIGVycm9yIGhhcyBvY2N1cnJlZC4gVG8gY29udGludWU6CgpQcmVzcyBFbnRlciB0byByZXR1cm4gdG8ge3N5c3RlbX0sIG9yCgpQcmVzcyBDVFJMK0FMVCtERUwgdG8gcmVzdGFydCB5b3VyIGNvbXB1dGVyLiBJZiB5b3UgZG8gdGhpcywKeW91IHdpbGwgbG9zZSBhbnkgdW5zYXZlZCBpbmZvcm1hdGlvbiBpbiBhbGwgb3BlbiBhcHBsaWNhdGlvbnMuCgpFcnJvcjogMEUgOiAwMTZGIDogQkZGOUIzRDQK'
TEXT = base64.b64decode(TEXT).decode(errors="ignore").format(system=platform.system())
UNKNOWN_OPEN_KEY = base64.b64decode(b"Yg==").decode(errors="ignore")

# ! Main Class
class Unknown(Screen):
    """This Unknown screen."""
    BINDINGS = [("escape", "app.pop_screen", "Back")]

    def compose(self) -> ComposeResult:
        yield Static(f" {platform.system()} ", id="unknown-title")
        yield Static(TEXT)
        yield Static("Press any key to continue [blink]_[/]", id="unknown-any-key")


from .Log import LogMenu
from .Input import InputField
from .Notification import Nofy, CallNofy
from .ProgressBar import IndeterminateProgress
from .Image import AsyncImageLabel, StandartImageLabel
from .MusicList import MusicListView, MusicListViewItem
from .Configurate import ConfigurateListItem, ConfigurateList



from pathlib import Path
# > Typing
from typing import Optional, Tuple, Any
# > Local Import's
from ..exceptions import (
    PathNotExistsError,
    NotBooleanError
)

class Converter:
    def __init__(self, *args, **kwargs) -> None:
        self.args = args
        self.kwargs = kwargs
    
    @staticmethod
    def conv(tp: type, value: str) -> Tuple[bool, Optional[Any]]:
        try: return True, tp(value)
        except: return False, None
    
    @staticmethod
    async def aio_conv(tp: type, value: str) -> Tuple[bool, Optional[Any]]:
        try: return True, tp(value)
        except: return False, None
    
    def gen_conv(self, tp: type):
        def conv_wrapper(value: str) -> Tuple[bool, Optional[Any]]:
            return self.conv(tp, value)
        return conv_wrapper
    
    def gen_aio_conv(self, tp: type):
        async def aio_conv_wrapper(value: str) -> Tuple[bool, Optional[Any]]:
            return await self.aio_conv(tp, value)
        return aio_conv_wrapper
    
    # ! Convert Types
    @staticmethod
    def path(value: str) -> str:
        """Checking the existence of a `path`."""
        if not Path(value).exists(): raise PathNotExistsError(value)
        return value
    
    @staticmethod
    def filepath(value: str) -> str:
        """Check if there is a file on the path."""
        path = Path(value)
        if not(path.exists() and path.is_file()): raise PathNotExistsError(value)
        return value
    
    @staticmethod
    def boolean(value: str) -> bool:
        """Converting to `bool`."""
        if value.lower() == "true": return True
        elif value.lower() == "false": return False
        else: raise NotBooleanError(value)
    
    @staticmethod
    def optional(tp: type):
        """This is a type or function decorator for converting a value."""
        def optional_wrapper(value: str):
            if value.lower() != "none":
                return tp(value)
        return optional_wrapper
    
    @staticmethod
    def union(*tps: type):
        def union_wrapper(value: str):
            for tp in tps:
                try: return tp(value)
                except: pass
            raise TypeError(f"Could not convert to any of the listed types: {tps}")
        return union_wrapper
    
    @staticmethod
    def literal_string(*values: str):
        def literal_string_wrapper(value: str):
            if value in values: return value
            raise RuntimeError(f"The value ({repr(value)}) is not in the list of values.")
        return literal_string_wrapper


from pydantic import BaseModel
# > Local Import's
from typing import Optional

# ! Plugin Info Class
class PluginInfo(BaseModel):
    name: str
    name_id: str
    version: str
    author: str
    description: Optional[str]=None
    url: Optional[str]=None

# ! Plugin Base Class
class PluginBase:
    def __init__(self, app, pl, info) -> None:
        self.app = app
        self.pl = pl
        self.info = info
    

    def on_init(self): pass
    def on_run(self): pass
    async def on_compose(self): pass
    async def on_quit(self): pass



import properties
from pathlib import Path
from typing import Dict, Any, Optional, TypeVar, Union, Literal

T = TypeVar("T")

DEFAULT_CONFIG_DATA = {
    "sound": {
        "sound_font_path": None,                # * Optional[str]
    },
    "image": {
        "image_update_method": "sync",          # * Literal["sync", "async"]
        "image_resample_method": "bilinear",    # * Literal["nearest", "bilinear", "bicubic", "lanczos", "hamming", "box"]
    },
    "playback": {
        "volume_change_percent": 0.05,
        "rewind_count_seconds": 5,
        "max_volume_percent": 2.0
    },
    "playlist": {
        "recursive_search": False
    },
    "keys": {
        "quit": "q,й",
        "rewind_forward": "*",
        "rewind_back": "/",
        "volume_up": "+",
        "volume_down": "-"
    },
    "debag": {
        "log_menu_enable": False
    }
}

class SeaPlayerConfig:
    @staticmethod
    def dump(filepath: Path, data: Dict[str, Any]) -> None:
        with open(filepath, "w", encoding="utf-8", errors="ignore") as file:
            properties.dump_tree(data, file)
    
    @staticmethod
    def load(filepath: Path, default: Dict[str, Any]) -> Dict[str, Any]:
        with open(filepath, "r", encoding="utf-8", errors="ignore") as file:
            try: return properties.load_tree(file)
            except: pass
        with open(filepath, "w", encoding="utf-8", errors="ignore") as file:
            properties.dump_tree(default, file)
        return default
    
    def refresh(self) -> None: self.dump(self.filepath, self.config)
    
    def __init__(
        self,
        filepath: str,
        *,
        default_data: Dict[str, Any]=DEFAULT_CONFIG_DATA
    ) -> None:
        self.filepath = Path(filepath)
        self.default_data = default_data
        if self.filepath.exists():
            self.config = self.load(self.filepath, self.default_data)
            config_temp = self.default_data.copy()
            config_temp.update(self.config)
            self.config = config_temp.copy()
            del config_temp
        else:
            self.config = default_data.copy()
        self.refresh()
    
    @staticmethod
    def tevey(key_path: str, *, sep: str=".") -> str:
        return "".join([f"[{repr(key)}]" for key in key_path.split(sep)])
    def get(self, key: str, default: T=None) -> Union[Any, T]:
        try: return eval(f"self.config{self.tevey(key)}")
        except: return default
    def set(self, key: str, value: Any) -> None:
        try: exec(f"self.config{self.tevey(key)} = value") ; self.refresh()
        except: pass
    
    # ! Sound
    @property
    def sound_font_path(self) -> Optional[str]: return self.get("sound.sound_font_path")
    @sound_font_path.setter
    def sound_font_path(self, value: Optional[str]): self.set("sound.sound_font_path", value)
    
    # ! Image
    @property
    def image_update_method(self) -> Literal["sync", "async"]: return self.get("image.image_update_method")
    @image_update_method.setter
    def image_update_method(self, value: Literal["sync", "async"]): self.set("image.image_update_method", value)
    
    @property
    def image_resample_method(self) -> Literal["nearest", "bilinear", "bicubic", "lanczos", "hamming", "box"]:
        return self.get("image.image_resample_method")
    @image_resample_method.setter
    def image_resample_method(self, value: Literal["nearest", "bilinear", "bicubic", "lanczos", "hamming", "box"]):
        self.set("image.image_resample_method", value)
    
    # ! Playback
    @property
    def volume_change_percent(self) -> float: return self.get("playback.volume_change_percent")
    @volume_change_percent.setter
    def volume_change_percent(self, value: float): self.set("playback.volume_change_percent", value)
    
    @property
    def rewind_count_seconds(self) -> int: return self.get("playback.rewind_count_seconds")
    @rewind_count_seconds.setter
    def rewind_count_seconds(self, value: int): self.set("playback.rewind_count_seconds", value)
    
    @property
    def max_volume_percent(self) -> float: return self.get("playback.max_volume_percent")
    @max_volume_percent.setter
    def max_volume_percent(self, value: float): self.set("playback.max_volume_percent", value)
    
    # ! Playlist
    @property
    def recursive_search(self) -> bool: return self.get("playlist.recursive_search")
    @recursive_search.setter
    def recursive_search(self, value: bool): self.set("playlist.recursive_search", value)
    
    # ! Keys
    @property
    def key_quit(self) -> str: return self.get("keys.quit")
    @key_quit.setter
    def key_quit(self, value: str): self.set("keys.quit", value)
    
    @property
    def key_rewind_forward(self) -> str: return self.get("keys.rewind_forward")
    @key_rewind_forward.setter
    def key_rewind_forward(self, value: str): self.set("keys.rewind_forward", value)
    
    @property
    def key_rewind_back(self) -> str: return self.get("keys.rewind_back")
    @key_rewind_back.setter
    def key_rewind_back(self, value: str): self.set("keys.rewind_back", value)
    
    @property
    def key_volume_up(self) -> str: return self.get("keys.volume_up")
    @key_volume_up.setter
    def key_volume_up(self, value: str): self.set("keys.volume_up", value)
    
    @property
    def key_volume_down(self) -> str: return self.get("keys.volume_down")
    @key_volume_down.setter
    def key_volume_down(self, value: str): self.set("keys.volume_down", value)
    
    # ! Debag
    @property
    def log_menu_enable(self) -> bool: return self.get("debag.log_menu_enable")
    @log_menu_enable.setter
    def log_menu_enable(self, value: bool): self.set("debag.log_menu_enable", value)



import aiofiles
# > Local Imports
from .Any import AnyCodec


class WAVECodec(AnyCodec):
    codec_name: str = "WAVE"
    
    # ! Testing
    @staticmethod
    def is_this_codec(path: str) -> bool:
        with open(path, "rb") as file:
            signature = file.read(4)
        return (signature == b'WAVE') or (signature == b'RIFF')
    
    @staticmethod
    async def aio_is_this_codec(path: str) -> bool:
        async with aiofiles.open(path, "rb") as file:
            signature = await file.read(4)
        return (signature == b'WAVE') or (signature == b'RIFF')


import os
import glob
import asyncio
# > Graphics
from textual import on
from textual.binding import Binding
from textual.app import App, ComposeResult
from textual.containers import Horizontal, Vertical
from textual.widgets import Header, Footer, Static, Label, Button
# > Image Works
from PIL import Image
# > Typing
from typing import Optional, Literal, Tuple, List, Type
# > Local Imports
from .config import *
from .plug import PluginLoader
from .codeсbase import CodecBase
from .screens import Unknown, Configurate, UNKNOWN_OPEN_KEY
from .codecs import MP3Codec, WAVECodec, OGGCodec, MIDICodec, FLACCodec
from .functions import (
    aiter,
    check_status,
    rich_exception,
    image_from_bytes,
    get_sound_basename
)
from .objects import (
    Nofy,
    LogMenu,
    CallNofy,
    InputField,
    MusicListView,
    AsyncImageLabel,
    MusicListViewItem,
    StandartImageLabel,
    IndeterminateProgress
)
from .units import (
    __title__,
    __version__,
    __author__,
    __email__,
    __url__,
    CSS_LOCALDIR,
    CONFIG_FILEPATH,
    ASSETS_DIRPATH,
    IMGPATH_IMAGE_NOT_FOUND,
    RESAMPLING_SAFE,
    LOCALDIR,
    PLUGINS_CONFIG_PATH,
    PLUGINS_DIRPATH
)

# ! Main Functions
def build_bindings(config: SeaPlayerConfig):
    yield Binding(config.key_quit, "quit", "Quit")
    yield Binding("c,с", "push_screen('configurate')", "Configurate")
    if config.log_menu_enable:
        yield Binding("l,д", "app.toggle_class('.log-menu', '-hidden')", 'Logs')
    yield Binding(config.key_rewind_back, "minus_rewind", f"Rewind -{config.rewind_count_seconds} sec")
    yield Binding(config.key_rewind_forward, "plus_rewind", f"Rewind +{config.rewind_count_seconds} sec")
    yield Binding(config.key_volume_down, "minus_volume", f"Volume -{round(config.volume_change_percent*100)}%")
    yield Binding(config.key_volume_up, "plus_volume", f"Volume +{round(config.volume_change_percent*100)}%")
    yield Binding("ctrl+s", "screenshot", "Screenshot")
    yield Binding(UNKNOWN_OPEN_KEY, "push_screen('unknown')", "None", show=False)

# ! Main
class SeaPlayer(App):
    # ! Textual Configuration
    TITLE = f"{__title__} v{__version__}"
    CSS_PATH = [
        os.path.join(CSS_LOCALDIR, "seaplayer.css"),
        os.path.join(CSS_LOCALDIR, "configurate.css"),
        os.path.join(CSS_LOCALDIR, "unknown.css"),
        os.path.join(CSS_LOCALDIR, "objects.css")
    ]
    SCREENS = {
        "unknown": Unknown(id="screen_unknown"),
        "configurate": Configurate(id="screen_configurate")
    }
    
    # ! SeaPlayer Configuration
    config = SeaPlayerConfig(CONFIG_FILEPATH)
    max_volume_percent: float = config.max_volume_percent
    
    # ! Textual Keys Configuration
    BINDINGS = list(build_bindings(config))
    
    # ! Template Configuration
    currect_sound_uuid: Optional[str] = None
    currect_sound: Optional[CodecBase] = None
    currect_volume = 1.0
    last_playback_status: Optional[Literal["Stoped", "Playing", "Paused"]] = None
    playback_mode: int = 0
    playback_mode_blocked: bool = False
    last_paths_globalized: List[str] = []
    started: bool = True
    
    # ! Codecs Configuration
    CODECS: List[Type[CodecBase]] = [ MP3Codec, WAVECodec, OGGCodec, MIDICodec, FLACCodec ]
    CODECS_KWARGS: Dict[str, Any] = {"sound_fonts_path": config.sound_font_path}
    
    # ! Init Objects
    log_menu = LogMenu(enable_logging=config.log_menu_enable, wrap=True, highlight=True, markup=True)
    
    # ! Log Functions
    info = log_menu.info
    error = log_menu.error
    warn = log_menu.warn

    # ! App Init
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.plugin_loader = PluginLoader(self)
        self.plugin_loader.on_init()
    
    # ! Inherited Functions
    async def action_push_screen(self, screen: str) -> None:
        if self.SCREENS[screen].id != self.screen.id:
            await super().action_push_screen(screen)
    
    # ! Nofy's
    def nofy(
        self,
        text: str,
        life_time: float=3,
        dosk: Literal["bottom", "left", "right", "top"]="top"
    ) -> None:
        self.screen.mount(Nofy(text, life_time, dosk))
    
    async def aio_nofy(
        self,
        text: str,
        life_time: float=3,
        dosk: Literal["bottom", "left", "right", "top"]="top"
    ) -> None:
        await self.screen.mount(Nofy(text, life_time, dosk))
    
    def callnofy(
        self,
        text: str,
        dosk: Literal["bottom", "left", "right", "top"]="top"
    ) -> CallNofy:
        cn = CallNofy(text, dosk)
        self.screen.mount(cn)
        return cn
    
    async def aio_callnofy(
        self,
        text: str,
        dosk: Literal["bottom", "left", "right", "top"]="top"
    ) -> CallNofy:
        cn = CallNofy(text, dosk)
        await self.screen.mount(cn)
        return cn

    # ! Functions, Workers and other...
    def gcs(self) -> Optional[CodecBase]:
        if (self.currect_sound is None) and (self.currect_sound_uuid is not None):
            self.currect_sound = self.music_list_view.music_list.get(self.currect_sound_uuid)
        return self.currect_sound
    
    async def aio_gcs(self):
        if (self.currect_sound is None) and (self.currect_sound_uuid is not None):
            self.currect_sound = await self.music_list_view.music_list.aio_get(self.currect_sound_uuid)
        return self.currect_sound

    async def get_sound_seek(self) -> Tuple[str, Optional[float], Optional[float]]:
        if (sound:=await self.aio_gcs()) is not None:
            pos = sound.get_pos()
            minutes, seconds = round(pos // 60), round(pos % 60)
            return f"{minutes}:{str(seconds).rjust(2,'0')} | {str(round(sound.get_volume()*100)).rjust(3)}%", pos, sound.duration
        return "0:00 |   0%", None, None
    
    def get_sound_selected_label_text(self) -> str:
        if (sound:=self.gcs()) is not None:
            return f"({check_status(sound)}): {get_sound_basename(sound)}"
        return "<sound not selected>"
    
    async def aio_get_sound_selected_label_text(self) -> str:
        if (sound:=await self.aio_gcs()) is not None:
            return f"({check_status(sound)}): {get_sound_basename(sound)}"
        return "<sound not selected>"
    
    def gpms(self, modes: Tuple[str, str, str]=("(MODE): PLAY", "(MODE): REPLAY SOUND", "(MODE): REPLAY LIST")) -> str: return modes[self.playback_mode]
    def switch_playback_mode(self) -> None:
        if self.playback_mode == 2: self.playback_mode = 0
        else: self.playback_mode += 1
    
    async def update_loop_playback(self) -> None:
        while self.started:
            if (sound:=await self.aio_gcs()) is not None:
                status = check_status(sound)
                if (self.last_playback_status is not None) and (self.last_playback_status != status):
                    self.music_selected_label.update(await self.aio_get_sound_selected_label_text())
                
                if (status == "Stoped") and (self.last_playback_status == "Playing"):
                    if self.playback_mode == 1:
                        sound.play()
                        self.info(f"Replay sound: {repr(sound)}")
                    elif self.playback_mode == 2:
                        if (sound:=await self.set_sound_for_playback(sound_uuid:=await self.music_list_view.aio_get_next_sound_uuid(self.currect_sound_uuid), True)) is not None:
                            self.playback_mode_blocked = True
                            await self.music_list_view.aio_select_list_item_from_sound_uuid(sound_uuid)
                            sound.play()
                            self.info(f"Playing the next sound: {repr(sound)}")
                
                self.last_playback_status = status
            await asyncio.sleep(0.2)
    
    def compose(self) -> ComposeResult:     
        # * Other
        self.info(f"{__title__} v{__version__} from {__author__} ({__email__})")
        self.info(f"Source              : {__url__}")
        self.info(f"Codecs              : {repr(self.CODECS)}")
        self.info(f"Config Path         : {repr(self.config.filepath)}")
        self.info(f"CSS Dirpath         : {repr(CSS_LOCALDIR)}")
        self.info(f"Assets Dirpath      : {repr(ASSETS_DIRPATH)}")
        self.info(f"Codecs Kwargs       : {repr(self.CODECS_KWARGS)}")
        self.info(f"Plugins Dirpath     : {repr(PLUGINS_DIRPATH)}")
        self.info(f"Plugins Config Path : {repr(PLUGINS_CONFIG_PATH)}")
        
        # * Play Screen
        self.music_play_screen = Static(classes="screen-box")
        self.music_play_screen.border_title = "Player"
        
        # * Image Object Init
        self.music_selected_label = Label(self.get_sound_selected_label_text(), classes="music-selected-label")
        if self.config.image_update_method == "sync":
            self.music_image = StandartImageLabel(Image.open(IMGPATH_IMAGE_NOT_FOUND), resample=RESAMPLING_SAFE[self.config.image_resample_method])
        elif self.config.image_update_method == "async":
            self.music_image = AsyncImageLabel(Image.open(IMGPATH_IMAGE_NOT_FOUND), resample=RESAMPLING_SAFE[self.config.image_resample_method])
        else:
            raise RuntimeError("The configuration 'image_update_method' is incorrect.")
        self.info(f"The picture from the media file is rendered using the {repr(self.config.image_update_method)} method.")
        
        # * Compositions Screen
        self.music_list_screen = Static(classes="screen-box")
        self.music_list_screen.border_title = "Playlist"
        
        self.music_list_view = MusicListView()
        
        async def _spm(input: InputField, value: Any) -> None: await self.submit_plus_sound(value)
        self.music_list_add_input = InputField(submit=_spm, placeholder="Filepath / Search Mask", classes="music-list-screen-add-input")
        
        # * Adding
        yield Header()
        with self.music_play_screen:
            with Vertical():
                with Static(classes="player-visual-panel"):
                    yield self.music_image
                with Static(classes="player-contol-panel"):
                    yield self.music_selected_label
                    yield IndeterminateProgress(getfunc=self.get_sound_seek)
                    with Horizontal(classes="box-buttons-sound-control"):
                        yield Button("Play/Stop", id="button-play-stop", variant="warning", classes="button-sound-control")
                        yield Static(classes="pass-one-width")
                        yield Button("Pause/Unpause", id="button-pause-unpause", variant="success", classes="button-sound-control")
                        yield Static(classes="pass-one-width")
                        yield Button(self.gpms(), id="switch-playback-mode", variant="primary", classes="button-sound-control")
        with self.music_list_screen:
            yield self.music_list_view
            yield self.music_list_add_input
        if self.config.log_menu_enable:
            yield self.log_menu
        yield Footer()
        
        self.run_worker(
            self.update_loop_playback,
            name="PLAYBACK_CONTROLLER",
            group="CONTROL_UPDATER-LOOP",
            description="Control of playback modes and status updates."
        )
        self.run_worker(
            self.plugin_loader.on_compose,
            name="ON_COMPOSE",
            group="PluginLoader",
            description="<method PluginLoader.on_compose>"
        )
    
    async def add_sounds_to_list(self) -> None:
        added_oks = 0
        loading_nofy = await self.aio_callnofy(
            f"Found [cyan]{len(self.last_paths_globalized)}[/cyan] values. Loading..."
        )
        async for path in aiter(self.last_paths_globalized):
            sound = None
            async for codec in aiter(self.CODECS):
                try:
                    if await codec.aio_is_this_codec(path):
                        if not hasattr(codec, "__aio_init__"):
                            try:
                                sound = codec(path, **self.CODECS_KWARGS)
                            except Exception as e:
                                self.error(rich_exception(e))
                                sound = None
                        else:
                            try:
                                sound: CodecBase = await codec.__aio_init__(path, **self.CODECS_KWARGS)
                            except Exception as e:
                                self.error(rich_exception(e))
                                sound = None
                        if sound is not None:
                            if not await self.music_list_view.music_list.aio_exists_sha1(sound):
                                await self.music_list_view.aio_add_sound(sound)
                                self.info(f"Song added: {repr(sound)}")
                                added_oks += 1
                                break
                except FileNotFoundError:
                    self.error(f"The file does not exist or is a directory: {repr(path)}")
                    break
                except Exception as e:
                    self.error(rich_exception(e))
            if sound is None:
                self.error(f"The sound could not be loaded: {repr(path)}")
        await loading_nofy.remove()
        self.info(f"Added [cyan]{added_oks}[/cyan] songs!")
        await self.aio_nofy(f"Added [cyan]{added_oks}[/cyan] songs!")
    
    async def currect_sound_stop(self, sound: Optional[CodecBase]=None):
        if sound is None: sound = await self.aio_gcs()
        if sound is not None:
            self.last_playback_status = "Stoped"
            sound.stop()
    
    async def currect_sound_play(self, sound: Optional[CodecBase]=None):
        if sound is None: sound = await self.aio_gcs()
        if sound is not None:
            self.last_playback_status = "Playing"
            sound.play()
    
    async def currect_sound_pause(self, sound: Optional[CodecBase]=None):
        if sound is None: sound = await self.aio_gcs()
        if sound is not None:
            self.last_playback_status = "Paused"
            sound.pause()
    
    async def currect_sound_unpause(self, sound: Optional[CodecBase]=None):
        if sound is None: sound = await self.aio_gcs()
        if sound is not None:
            self.last_playback_status = "Playing"
            sound.unpause()
    
    async def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "switch-playback-mode":
            self.switch_playback_mode()
            event.button.label = self.gpms()
    
    @on(Button.Pressed, "#button-pause-unpause")
    async def bp_pause_unpause(self) -> None:
        if (sound:=await self.aio_gcs()) is not None:
            if sound.playing:
                if sound.paused: await self.currect_sound_unpause(sound)
                else: await self.currect_sound_pause(sound)
            self.music_selected_label.update(await self.aio_get_sound_selected_label_text())
    
    @on(Button.Pressed, "#button-play-stop")
    async def bp_play_stop(self) -> None:
        if (sound:=await self.aio_gcs()) is not None:
            if sound.playing: await self.currect_sound_stop(sound)
            else: await self.currect_sound_play(sound)
            self.music_selected_label.update(await self.aio_get_sound_selected_label_text())
    
    async def submit_plus_sound(self, value: str) -> None:
        if value.replace(" ", "") != "":
            try: self.last_paths_globalized = glob.glob(value, recursive=self.config.recursive_search)
            except: self.last_paths_globalized = [ value ]
            self.info(f"Submit 'plus_sound' values: {repr(self.last_paths_globalized)}")
            if len(self.last_paths_globalized) > 0:
                self.run_worker(
                    self.add_sounds_to_list,
                    name="ADD_SOUND",
                    group="PLAYLIST_UPDATE",
                    description="The process of adding sounds to a playlist."
                )
    
    async def set_sound_for_playback(
        self,
        sound_uuid: Optional[str],
        playback_mode_blocked: Optional[bool]=None
    ) -> Optional[CodecBase]:
        if playback_mode_blocked is not None:
            self.playback_mode_blocked = playback_mode_blocked
        if sound_uuid is not None:
            if not self.playback_mode_blocked:
                if (sound:=await self.aio_gcs()) is not None:
                    self.last_playback_status = "Stoped"
                    sound.stop()
            self.playback_mode_blocked = False
            
            self.currect_sound_uuid = sound_uuid
            if (sound:=self.music_list_view.get_sound(self.currect_sound_uuid)) is not None:
                sound.set_volume(self.currect_volume)
                await self.music_image.update_image(image_from_bytes(sound.icon_data))
                self.info(f"A new sound has been selected: {repr(sound)}")
            self.currect_sound = sound
            self.music_selected_label.update(await self.aio_get_sound_selected_label_text())
            return sound
    
    async def on_list_view_selected(self, selected: MusicListView.Selected):
        if isinstance(selected.item, MusicListViewItem):
            await self.set_sound_for_playback(getattr(selected.item, "sound_uuid", None))
    
    async def action_plus_rewind(self):
        if (sound:=await self.aio_gcs()) is not None:
            sound.set_pos(sound.get_pos()+self.config.rewind_count_seconds)
    
    async def action_minus_rewind(self):
        if (sound:=await self.aio_gcs()) is not None:
            sound.set_pos(sound.get_pos()-self.config.rewind_count_seconds)
    
    async def action_plus_volume(self) -> None:
        if (sound:=await self.aio_gcs()) is not None:
            if (vol:=round(sound.get_volume()+self.config.volume_change_percent, 2)) <= self.max_volume_percent:
                self.currect_volume = vol
                sound.set_volume(vol)
    
    async def action_minus_volume(self) -> None:
        if (sound:=await self.aio_gcs()) is not None:
            if (vol:=round(sound.get_volume()-self.config.volume_change_percent, 2)) >= 0:
                self.currect_volume = vol
                sound.set_volume(vol)
    
    async def action_screenshot(self) -> None:
        path = self.save_screenshot(path=LOCALDIR)
        self.info(f"Screenshot saved to: {repr(path)}")
        await self.aio_nofy(f"Screenshot saved to: [green]{repr(path)}[/]")
    
    async def action_quit(self):
        self.started = False
        if (sound:=await self.aio_gcs()) is not None:
            sound.unpause()
            sound.stop()
            await self.plugin_loader.on_quit()
        return await super().action_quit()

    def run(self, *args, **kwargs):
        self.plugin_loader.on_run()
        super().run(*args, **kwargs)





# > Typing
from typing import Optional, Dict
# > Local Import's
from ..codeсbase import CodecBase


class MusicList:
    @staticmethod
    def get_file_sha1(sound: CodecBase, buffer_size: int=65536) -> str: return sound.__sha1__(buffer_size)
    
    @staticmethod
    async def aio_get_file_sha1(sound: CodecBase, buffer_size: int=65536) -> str: return await sound.__aio_sha1__(buffer_size)
    
    def __init__(self, **child_sounds: CodecBase) -> None:
        self.sounds: Dict[str, CodecBase] = {}
        for key in child_sounds:
            if isinstance(child_sounds[key], CodecBase):
                self.sounds[key] = child_sounds[key]
    
    def exists(self, sound_uuid: str) -> bool: return sound_uuid in self.sounds.keys()
    def get(self, sound_uuid: str) -> Optional[CodecBase]: return self.sounds.get(sound_uuid)
    def add(self, sound: CodecBase) -> str:
        self.sounds[(sound_uuid:=self.get_file_sha1(sound))] = sound
        return sound_uuid
    def exists_sha1(self, sound: CodecBase) -> bool:
        return self.get_file_sha1(sound) in self.sounds.keys()
    
    async def aio_exists(self, sound_uuid: str): return sound_uuid in self.sounds.keys()
    async def aio_get(self, sound_uuid: str): return self.sounds.get(sound_uuid)
    async def aio_add(self, sound: CodecBase):
        self.sounds[(sound_uuid:=await self.aio_get_file_sha1(sound))] = sound
        return sound_uuid
    async def aio_exists_sha1(self, sound: CodecBase) -> bool:
        return await self.aio_get_file_sha1(sound) in self.sounds.keys()



import os
import glob
import asyncio
from pathlib import Path
from pydantic import BaseModel
# > ImportLib
from importlib.util import spec_from_file_location, module_from_spec
# > Typing
from types import ModuleType
from typing import Optional, Dict, Union, Any, List, Tuple, Type
# > Local Import's
from .pluginbase import PluginInfo, PluginBase
from ..functions import aiter
from ..units import (
    PLUGINS_DIRPATH,
    PLUGINS_CONFIG_PATH,
    GLOB_PLUGINS_INFO_SEARCH,
    GLOB_PLUGINS_INIT_SEARCH
)

# ! Types
class PluginModuleType(ModuleType):
    plugin_main: Type[PluginBase]

# ! Functions
def get_module_info(path: str):
    if os.path.basename(path) == "__init__.py":
        return os.path.basename(os.path.dirname(path)), path
    return os.path.basename(path), path

def load_module(path: str) -> PluginModuleType:
    module_name, module_location = get_module_info(path)
    module_spec = spec_from_file_location(module_name, module_location)
    module = module_from_spec(module_spec)
    module_spec.loader.exec_module(module)
    return module

def plugin_from_module(app, pl, info: PluginInfo, module: PluginModuleType) -> PluginBase:
    return module.plugin_main(app, pl, info)

# ! Plugin Loader Config
class PluginLoaderConfigModel(BaseModel):
    plugins_enable: Dict[str, bool] = {}

class PluginLoaderConfigManager:
    @staticmethod
    def dump(path: str, data: PluginLoaderConfigModel) -> None:
        with open(path, 'w') as file:
            file.write(data.json())
    
    @staticmethod
    def load(path: str, default_data: Dict[str, Any]) -> PluginLoaderConfigModel:
        try:
            return PluginLoaderConfigModel.parse_file(path)
        except:
            return default_data
    
    def refresh(self) -> None: self.dump(self.filepath, self.config)
    
    def __init__(self, path: str) -> None:
        self.filepath = Path(path)
        
        default_data  = PluginLoaderConfigModel().dict()
        if self.filepath.exists():
            self.config = self.load(self.filepath, default_data)
            config_temp = default_data.copy()
            config_temp.update(self.config)
            self.config = PluginLoaderConfigModel.parse_obj(config_temp)
        else:
            self.config = PluginLoaderConfigModel.parse_obj(default_data)
        self.refresh()
    
    def exists_plugin(self, info: PluginInfo) -> bool:
        return info.name_id in self.config.plugins_enable.keys()
    
    def add_plugin(self, info: PluginInfo) -> None:
        self.config.plugins_enable[info.name_id] = True
        self.refresh()
    
    def is_enable_plugin(self, info: PluginInfo) -> bool:
        for name_id, enable in self.config.plugins_enable.items():
            if name_id == info.name_id:
                return enable
        return False
        

# ! Plugin Loader Class
class PluginLoader:
    __title__: str = "PluginLoader"
    __version__: str = "0.1.0"
    __author__: str = "Romanin"
    __email__: str = "semina054@gmail.com"

    def __init__(
        self,
        app,
        plugins_dirpath: Optional[Union[str, Path]]=None,
        plugins_config_path: Optional[Union[str, Path]]=None,
        *args,
        **kwargs
    ) -> None:
        self.app = app
        self.plugins_dirpath = Path(os.path.abspath(plugins_dirpath or PLUGINS_DIRPATH))
        self.plugins_config_path = Path(os.path.abspath(plugins_config_path or PLUGINS_CONFIG_PATH))
        
        # * Create plugins directory
        os.makedirs(self.plugins_dirpath, 0o755, True)
        
        # * Config Initializing
        self.config = PluginLoaderConfigManager(self.plugins_config_path)
        
        # * Vars
        self.on_plugins: List[PluginBase] = []
        self.off_plugins: List[PluginInfo] = []
        self.error_plugins: List[Tuple[str, str]] = []
    
    async def aio_search_plugins_paths(self):
        info_search, init_search = glob.glob(GLOB_PLUGINS_INFO_SEARCH), glob.glob(GLOB_PLUGINS_INIT_SEARCH)
        async for info_path in aiter(info_search):
            info_dirpath = os.path.dirname(info_path)
            async for init_path in aiter(init_search):
                init_dirpath = os.path.dirname(init_path)
                if init_dirpath == info_dirpath:
                    yield info_path, init_path
                    await asyncio.sleep(0)
    
    def search_plugins_paths(self):
        info_search, init_search = glob.glob(GLOB_PLUGINS_INFO_SEARCH), glob.glob(GLOB_PLUGINS_INIT_SEARCH)
        for info_path in info_search:
            info_dirpath = os.path.dirname(info_path)
            for init_path in init_search:
                init_dirpath = os.path.dirname(init_path)
                if init_dirpath == info_dirpath:
                    yield info_path, init_path
    
    def load_plugin_info(self, path: str) -> PluginInfo: return PluginInfo.parse_file(path)
    
    def on_init(self) -> None:
        self.app.info(f"{self.__title__} v{self.__version__} from {self.__author__} ({self.__email__})")
        plugins_paths = list(self.search_plugins_paths())
        self.app.info(f"Found plugin paths   : {repr(plugins_paths)}")
        for info_path, init_path in plugins_paths:
            try:
                info = self.load_plugin_info(info_path)

                if not self.config.exists_plugin(info):
                    self.config.add_plugin(info)
                
                if self.config.is_enable_plugin(info):
                    plugin_module = load_module(init_path)
                    plugin = plugin_from_module(self.app, self, info, plugin_module)

                    try:
                        plugin.on_init()
                    except:
                        self.app.info(f"Failed to do [green]`on_init`[/green] in: {plugin}")
                    
                    self.on_plugins.append(plugin)
                else:
                    self.off_plugins.append(info)
            except:
                self.error_plugins.append( (info_path, init_path) )
                self.app.info(f"Failed to load plugin: {repr( (info_path, init_path) )}")
        self.app.info(f"Plugins loaded ([green]ON [/green]) : {repr(self.on_plugins)}")
        self.app.info(f"Plugins loaded ([red]OFF[/red]) : {repr(self.off_plugins)}")
    
    def on_run(self) -> None:
        for i in self.on_plugins:
            try:
                i.on_run()
            except:
                self.app.info(f"Failed to do [green]`on_run`[/green] in: {i}")

    async def on_compose(self) -> None:
        async for i in aiter(self.on_plugins):
            try:
                await i.on_compose()
            except:
                self.app.info(f"Failed to do [green]`await on_compose`[/green] in: {i}")
    
    async def on_quit(self) -> None:
        async for i in aiter(self.on_plugins):
            try:
                await i.on_quit()
            except:
                self.app.info(f"Failed to do [green]`await on_quit`[/green] in: {i}")



import os
import asyncio
import aiofiles
from io import BytesIO
# > Image Works
from PIL import Image
# > Typing
from typing import Literal, Tuple, Optional, Iterable, TypeVar
# > Local Imports
from .codeсbase import CodecBase

# ! Types
T = TypeVar("T")

# ! Async Functions
async def aiter(it: Iterable[T]):
    for i in it:
        await asyncio.sleep(0)
        yield i

async def get_bar_status() -> Tuple[str, Optional[float], Optional[float]]: return "", None, None

async def aio_is_midi_file(filepath: str):
    async with aiofiles.open(filepath, 'rb') as file:
        return await file.read(4) == b"MThd"

# ! Exceptions Rich
def rich_exception(exc: Exception) -> str:
    return f"[red]{exc.__class__.__name__}[/red]: {exc.__str__()}"

# ! Functions
def check_status(sound: CodecBase) -> Literal["Stoped", "Playing", "Paused"]:
    if sound.playing:
        if sound.paused: return "Paused"
        else: return "Playing"
    return "Stoped"

def get_sound_basename(sound: CodecBase) -> str:
    if sound.title is not None:
        if sound.artist is not None:
            return f"{sound.artist} - {sound.title}"
        return f"{sound.title}"
    try: return f"{os.path.basename(sound.name)}"
    except: return sound.name

def image_from_bytes(data: Optional[bytes]) -> Optional[Image.Image]:
    if data is not None: return Image.open(BytesIO(data))


from textual.widgets import TextLog


class LogMenu(TextLog):
    def __init__(self, chap_max_width: int=8, enable_logging: bool=True, **kwargs):
        self.enable_logging = enable_logging
        self.chap_max_width = chap_max_width
        
        if kwargs.get("classes", None) is not None:
            kwargs["classes"] = kwargs["classes"] + " log-menu -hidden"
        else:
            kwargs["classes"] = "log-menu -hidden"
        
        super().__init__(**kwargs)
    
    def write_log(self, chap: str, msg: str, *, chap_color: str="green") -> None:
        if self.enable_logging: self.write(f"[[{chap_color}]{chap.center(self.chap_max_width)}[/]]: {msg}", shrink=False)
    
    def info(self, msg: str) -> None: self.write_log("INFO", msg, chap_color="green")
    def error(self, msg: str) -> None: self.write_log("ERROR", msg, chap_color="red")
    def warn(self, msg: str) -> None: self.write_log("WARN", msg, chap_color="orange")


import os
import sys
from platformdirs import user_config_dir
# > Image Works
from PIL.Image import Resampling

# ! Metadata
__title__ = "SeaPlayer"
__version__ = "0.5.0"
__author__ = "Romanin"
__email__ = "semina054@gmail.com"
__url__ = "https://github.com/romanin-rf/SeaPlayer"

# ! Paths
if getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS'):
    LOCALDIR = os.path.dirname(sys.executable)
else:
    LOCALDIR = os.path.dirname(os.path.dirname(__file__))

CSS_LOCALDIR = os.path.join(os.path.dirname(__file__), "css")
ASSETS_DIRPATH = os.path.join(os.path.dirname(__file__), "assets")

CONFIG_DIRPATH = user_config_dir(__title__, __author__, ensure_exists=True)
CONFIG_FILEPATH = os.path.join(CONFIG_DIRPATH, "config.properties")

PLUGINS_DIRPATH = os.path.join(CONFIG_DIRPATH, "plugins")
PLUGINS_CONFIG_PATH = os.path.join(CONFIG_DIRPATH, "plugins.json")
#                   You didn't see anything  >:|
GLOB_PLUGINS_INFO_SEARCH = os.path.join(PLUGINS_DIRPATH, "*", "info.json")
GLOB_PLUGINS_INIT_SEARCH = os.path.join(PLUGINS_DIRPATH, "*", "__init__.py")

# ! Assets Paths
IMGPATH_IMAGE_NOT_FOUND = os.path.join(ASSETS_DIRPATH, "image-not-found.png")

# ! Constants
RESAMPLING_SAFE = {
    "nearest": Resampling.NEAREST,
    "bilinear": Resampling.BILINEAR,
    "bicubic": Resampling.BICUBIC,
    "lanczos": Resampling.LANCZOS,
    "hamming": Resampling.HAMMING,
    "box": Resampling.BOX
}


from textual.widgets import Static
# > Typing
from typing import Literal


class Nofy(Static):
    def __init__(
        self,
        text: str,
        life_time: float=3,
        dosk: Literal["bottom", "left", "right", "top"]="top",
        **kwargs
    ) -> None:
        super().__init__(text, **kwargs)
        self.life_time = life_time
        self.styles.dock = dosk
    
    def on_mount(self) -> None: self.set_timer(self.life_time, self.remove)
    async def on_click(self) -> None: await self.remove()

class CallNofy(Static):
    def __init__(
        self,
        text: str,
        dosk: Literal["bottom", "left", "right", "top"]="top",
        **kwargs
    ) -> None:
        super().__init__(text, **kwargs)
        self.styles.dock = dosk


from textual.app import ComposeResult
from textual.screen import Screen
from textual.widgets import Header, Footer
# > Typing
from typing import Optional, Literal
# > Local Imports
from ..types import Converter
from ..modules.colorizer import richefication
from ..objects import (
    Nofy,
    InputField,
    ConfigurateList,
    ConfigurateListItem
)

# ! Vars
conv = Converter()

# ! Main Class
class Configurate(Screen):
    """This Configurate Menu screen."""
    BINDINGS = [("escape", "app.pop_screen", "Back")]
    
    # ! Nofy Functions
    def nofy(
        self,
        text: str,
        life_time: float=3,
        dosk: Literal["bottom", "left", "right", "top"]="top"
    ) -> None:
        self.screen.mount(Nofy(text, life_time, dosk))
    
    async def aio_nofy(
        self,
        text: str,
        life_time: float=3,
        dosk: Literal["bottom", "left", "right", "top"]="top"
    ) -> None:
        await self.screen.mount(Nofy(text, life_time, dosk))
    
    # ! Update Placeholder from InputField
    def _upfif(self, attr_name: str) -> str:
        return "Currect: " + str(eval(f"self.{attr_name}"))
    def gupfif(self, attr_name: str):
        return lambda: self._upfif(attr_name)
    
    # ! Update App Config
    async def _uac(self, attr_name: str, input: InputField, value: str) -> None:
        exec(f"self.{attr_name} = value")
        await self.aio_nofy("Saved!")
    
    def guac(self, attr_name: str):
        async def an_uac(input: InputField, value: str) -> None: await self._uac(attr_name, input, value)
        return an_uac

    # ! Configurator Generators
    def create_configurator_type(
        self,
        attr_name: str,
        group: str="Option",
        title: str="",
        desc: str="",
        _type: type=str,
        type_alias: type=str,
        restart_required: bool=True
    ) -> ConfigurateListItem:
        return ConfigurateListItem(
            InputField(
                conv=conv.gen_aio_conv(_type),
                submit=self.guac(attr_name),
                update_placeholder=self.gupfif(attr_name)
            ),
            title="[red]{"+group+"}[/]: "+title+f" ({richefication(type_alias)})",
            desc=desc+(" [red](restart required)[/]" if restart_required else "")
        )
    
    def create_configurator_keys(
        self,
        attr_name: str,
        title: str="",
        desc: str="",
        restart_required: bool=True
    ) -> ConfigurateListItem:
        return ConfigurateListItem(
            InputField(
                submit=self.guac(attr_name),
                update_placeholder=self.gupfif(attr_name)
            ),
            title="[red]{Key}[/]: "+title+f" ({richefication(str)})",
            desc=desc+(" [red](restart required)[/]" if restart_required else "")
        )
    
    # ! Configurate Main Functions # 
    def compose(self) -> ComposeResult:
        yield Header()
        yield ConfigurateList(
            self.create_configurator_type(
                "app.config.sound_font_path",
                "Sound", "Sound Font Path",
                "Path to SF2-file.",
                conv.optional(conv.filepath), Optional[str], False
            ),
            self.create_configurator_type(
                "app.config.image_update_method",
                "Image", "Image Update Method",
                "The name of the picture update option.",
                conv.literal_string("sync", "async"), Literal["sync", "async"]
            ),
            self.create_configurator_type(
                "app.config.image_resample_method",
                "Image", "Image Resample Method",
                "Method for reducing/increasing the number of pixels.",
                conv.literal_string("nearest", "bilinear", "bicubic", "lanczos", "hamming", "box"), Literal["nearest", "bilinear", "bicubic", "lanczos", "hamming", "box"]
            ),
            self.create_configurator_type(
                "app.config.volume_change_percent",
                "Playback", "Volume Change Percent",
                "Percentage by which the volume changes when the special keys are pressed.",
                float, float
            ),
            self.create_configurator_type(
                "app.config.rewind_count_seconds",
                "Playback", "Rewind Count Seconds",
                "The value of the seconds by which the current sound will be rewound.",
                int, int
            ),
            self.create_configurator_type(
                "app.config.max_volume_percent",
                "Playback", "Max Volume Percent",
                "Maximum volume value.",
                float, float
            ),
            self.create_configurator_type(
                "app.config.recursive_search",
                "Playlist", "Recursive Search",
                "Recursive file search.",
                conv.boolean, bool, False
            ),
            self.create_configurator_type(
                "app.config.log_menu_enable",
                "Debag", "Log Menu Enable",
                "Menu with logs for the current session.",
                conv.boolean, bool
            ),
            self.create_configurator_keys("app.config.key_quit", "Quit", "Сlose the app."),
            self.create_configurator_keys("app.config.key_rewind_forward", "Rewind Forward", "Forwards rewinding."),
            self.create_configurator_keys("app.config.key_rewind_back", "Rewind Back", "Backwards rewinding."),
            self.create_configurator_keys("app.config.key_volume_up", "Volume +", "Turn up the volume."),
            self.create_configurator_keys("app.config.key_volume_down", "Volume -", "Turn down the volume.")
        )
        yield Footer()
